
LOGGING_STR = "logging"
CONSTELLATIONS_STR = 'constellations'
SIGNALS_STR = 'signals'
RATE_STR = 'rate'
MEAS_PROP_STR = 'measurement_properties'
SOLUTION_STR = 'solution'
MEASUREMENTS_STR = 'measurements'
EPHEMERIS_STR = 'ephemeris'
ENABLED_STR = "enabled"
SMOOTHING_STR = "smoothing"
FILE_ROTATION_STR = "file_rotation"

GPS_STR = 'GPS'
GAL_STR = 'GAL'
BDS_STR = 'BDS'
QZSS_STR = 'QZSS'
GLO_STR = 'GLO'
